(*Based on Bertram Felgenhauer's Haskell code of
haskell-rewriting/union-find-array*)
open Prelude;;

module UF = UnionFind;;

module type ANNOTATION = sig
  type t
  val default : t
end

module type SIGNATURE = sig
  type node = int
  type annot
  include Monad.STATE_MONAD
  val new_node : annot -> node t
  val lookup : node -> (node * annot) t
  val merge : (annot -> annot -> annot) -> node -> node -> unit t
  val annotate : node -> annot -> unit t
  val run' : 'a t ->  annot UF.t * 'a
end

module Make (A : ANNOTATION) : SIGNATURE with
  type annot = A.t =
struct

type node = int
type annot = A.t

module UState = struct
  (*state consisting of next fresh node and a annotated union-find data structure*)
  type t = { next : int; forest : A.t UF.t }
end

include Monad.State (UState);;

let new_node l =
  get >>= fun u ->
  let size = UF.size u.UState.forest in
  let n = u.UState.next in
  begin if size <= n then begin
    let forest' = UF.grow u.UState.forest (2 * size) in
    UF.annotate forest' n l;
    set { UState.forest = forest'; UState.next = n + 1 }
  end else begin
    UF.annotate u.UState.forest n l;
    set { u with UState.next = n + 1 }
  end end >>
  return n
;;

let lookup n =
  get >>= fun u ->
  return (UF.lookup u.UState.forest n)
;;

let merge f i j =
  get >>= fun u ->
  ignore (UF.merge u.UState.forest (fun x y -> (f x y, ())) i j);
  return ()
;;

let annotate n l =
  get >>= fun u ->
  UF.annotate u.UState.forest n l;
  return ()
;;

let run' a =
  let (x, s) = eval { UState.next = 0; UState.forest = UF.mk 8 A.default } a in
  (s.UState.forest, x)
;;

end

